@@ -40,7 +40,15 @@ function terror(e)
   error(e)
 end
 
-version = 0.9.2
+function ttable_size(t)
+  local count = 0
+  for k, v in pairs(t) do
+    count = count + 1
+  end
+  return count
+end
+
+version = 0.9.3
 
 TType = {
   STOP   = 0,
@@ -69,7 +77,7 @@ TMessageType = {
   ONEWAY = 4
 }
 
--- Recursive __index function to achive inheritance
+-- Recursive __index function to achieve inheritance
 function __tobj_index(self, key)
   local v = rawget(self, key)
   if v ~= nil then
